@@ -73,6 +73,7 @@ const NUMBER_OF_SECTIONS_IN_SEARCH = 2;
 let wideScreen = false;
 let displayGridColumnWidth = null;
 let displayGridRowHeight = null;
+let addRemoveButtonSize = null;
 
 function Overview() {
     this._init();
@@ -80,10 +81,6 @@ function Overview() {
 
 Overview.prototype = {
     _init : function() {
-        let me = this;
-
-        let global = Shell.Global.get();
-
         this._group = new Clutter.Group();
         this._group._delegate = this;
 
@@ -95,7 +92,15 @@ Overview.prototype = {
 
         this._activeDisplayPane = null;
 
-        // Used to catch any clicks when we have an active pane; see the comments
+        // During transitions, we raise this to the top to avoid having the overview
+        // area be reactive; it causes too many issues such as double clicks on
+        // Dash elements, or mouseover handlers in the workspaces.
+        this._coverPane = new Clutter.Rectangle({ opacity: 0,
+                                                  reactive: true });
+        this._group.add_actor(this._coverPane);
+        this._coverPane.connect('event', Lang.bind(this, function (actor, event) { return true; }));
+
+        // Similar to the cover pane but used for dialogs ("panes"); see the comments
         // in addPane below.
         this._transparentBackground = new Clutter.Rectangle({ opacity: 0,
                                                               reactive: true });
@@ -127,12 +132,12 @@ Overview.prototype = {
         this._transparentBackground.lower_bottom();
         this._paneContainer.lower_bottom();
 
+        this._coverPane.lower_bottom();
+
         this._workspaces = null;
     },
 
     _recalculateGridSizes: function () {
-        let global = Shell.Global.get();
-
         wideScreen = (global.screen_width/global.screen_height > WIDE_SCREEN_CUT_OFF_RATIO);
 
         // We divide the screen into an imaginary grid which helps us determine the layout of
@@ -147,12 +152,14 @@ Overview.prototype = {
     },
 
     relayout: function () {
-        let global = Shell.Global.get();
-
         let screenHeight = global.screen_height;
         let screenWidth = global.screen_width;
 
-        let contentHeight = screenHeight - Panel.PANEL_HEIGHT;
+        let contentY = Panel.PANEL_HEIGHT;
+        let contentHeight = screenHeight - contentY;
+
+        this._coverPane.set_position(0, contentY);
+        this._coverPane.set_size(screenWidth, contentHeight);
 
         let workspaceColumnsUsed = wideScreen ? COLUMNS_FOR_WORKSPACES_WIDE_SCREEN : COLUMNS_FOR_WORKSPACES_REGULAR_SCREEN;
         let workspaceRowsUsed = wideScreen ? ROWS_FOR_WORKSPACES_WIDE_SCREEN : ROWS_FOR_WORKSPACES_REGULAR_SCREEN;
@@ -167,22 +174,21 @@ Overview.prototype = {
         this._workspacesX = displayGridColumnWidth + WORKSPACE_GRID_PADDING;
         this._workspacesY = displayGridRowHeight + WORKSPACE_GRID_PADDING * (screenHeight / screenWidth);
 
-        let dashY = Panel.PANEL_HEIGHT;
-        this._dash.actor.set_position(0, dashY);
-        this._dash.actor.set_size(displayGridColumnWidth, screenHeight - dashY);
-        this._dash.searchArea.height = this._workspacesY - dashY;
+        this._dash.actor.set_position(0, contentY);
+        this._dash.actor.set_size(displayGridColumnWidth, contentHeight);
+        this._dash.searchArea.height = this._workspacesY - contentY;
         this._dash.sectionArea.height = this._workspacesHeight;
 
         // place the 'Add Workspace' button in the bottom row of the grid
-        this._addButtonSize = Math.floor(displayGridRowHeight * 3/5);
-        this._addButtonX = this._workspacesX + this._workspacesWidth - this._addButtonSize;
+        addRemoveButtonSize = Math.floor(displayGridRowHeight * 3/5);
+        this._addButtonX = this._workspacesX + this._workspacesWidth - addRemoveButtonSize;
         this._addButtonY = screenHeight - Math.floor(displayGridRowHeight * 4/5);
 
-        this._backOver.set_position(0, Panel.PANEL_HEIGHT);
+        this._backOver.set_position(0, contentY);
         this._backOver.set_size(global.screen_width, contentHeight);
 
         this._paneContainer.set_position(this._dash.actor.x + this._dash.actor.width + DEFAULT_PADDING,
-                                         Panel.PANEL_HEIGHT);
+                                         contentY);
         // Dynamic width
         this._paneContainer.height = contentHeight;
 
@@ -232,7 +238,7 @@ Overview.prototype = {
     // This allows the user to place the item on any workspace.
     handleDragOver : function(source, actor, x, y, time) {
         if (source instanceof GenericDisplay.GenericDisplayItem
-            || source instanceof AppDisplay.WellDisplayItem) {
+            || source instanceof AppDisplay.BaseWellItem) {
             if (this._activeDisplayPane != null)
                 this._activeDisplayPane.close();
             return true;
@@ -271,20 +277,16 @@ Overview.prototype = {
     show : function() {
         if (this.visible)
             return;
-        if (!Main.beginModal())
-            return;
+        Main.pushModal(this._dash.actor);
 
         this.visible = true;
         this.animationInProgress = true;
 
-        let global = Shell.Global.get();
-
         this._dash.show();
 
         /* TODO: make this stuff dynamic */
         this._workspaces = new Workspaces.Workspaces(this._workspacesWidth, this._workspacesHeight,
-                                                     this._workspacesX, this._workspacesY,
-                                                     this._addButtonSize, this._addButtonX, this._addButtonY);
+                                                     this._workspacesX, this._workspacesY);
         this._group.add_actor(this._workspaces.actor);
 
         // The workspaces actor is as big as the screen, so we have to raise the dash above it
@@ -292,6 +294,12 @@ Overview.prototype = {
         // be as big as the screen.
         this._dash.actor.raise(this._workspaces.actor);
 
+        // Create (+) button
+        this._addButton = new AddWorkspaceButton(addRemoveButtonSize, this._addButtonX, this._addButtonY, Lang.bind(this, this._acceptNewWorkspaceDrop));
+        this._addButton.actor.connect('button-release-event', Lang.bind(this, this._addNewWorkspace));
+        this._group.add_actor(this._addButton.actor);
+        this._addButton.actor.raise(this._workspaces.actor);
+
         // All the the actors in the window group are completely obscured,
         // hiding the group holding them while the Overview is displayed greatly
         // increases performance of the Overview especially when there are many
@@ -327,15 +335,14 @@ Overview.prototype = {
                            time: ANIMATION_TIME
                          });
 
+        this._coverPane.raise_top();
         this.emit('showing');
     },
 
-    hide : function() {
+    hide: function() {
         if (!this.visible || this._hideInProgress)
             return;
 
-        let global = Shell.Global.get();
-
         this.animationInProgress = true;
         this._hideInProgress = true;
         if (this._activeDisplayPane != null)
@@ -365,6 +372,7 @@ Overview.prototype = {
                            time: ANIMATION_TIME
                          });
 
+        this._coverPane.raise_top();
         this.emit('hiding');
     },
 
@@ -375,6 +383,18 @@ Overview.prototype = {
             this.show();
     },
 
+    /**
+     * getWorkspacesForWindow:
+     * @metaWindow: A #MetaWindow
+     *
+     * Returns the Workspaces object associated with the given window.
+     * This method is not be accessible if the overview is not open
+     * and will return %null.
+     */
+    getWorkspacesForWindow: function(metaWindow) {
+        return this._workspaces;
+    },
+
     /**
      * activateWindow:
      * @metaWindow: A #MetaWindow
@@ -387,7 +407,6 @@ Overview.prototype = {
      */
     activateWindow: function (metaWindow, time) {
          this._workspaces.activateWindowFromOverview(metaWindow, time);
-         this.hide();
     },
 
     //// Private methods ////
@@ -397,13 +416,12 @@ Overview.prototype = {
             return;
 
         this.animationInProgress = false;
+        this._coverPane.lower_bottom();
 
         this.emit('shown');
     },
 
     _hideDone: function() {
-        let global = Shell.Global.get();
-
         global.window_group.show();
 
         this._workspaces.destroy();
@@ -416,8 +434,47 @@ Overview.prototype = {
         this.animationInProgress = false;
         this._hideInProgress = false;
 
-        Main.endModal();
+        this._coverPane.lower_bottom();
+
+        Main.popModal(this._dash.actor);
         this.emit('hidden');
+    },
+
+    _addNewWorkspace: function() {
+        global.screen.append_new_workspace(false, global.screen.get_display().get_current_time());
+    },
+
+    _acceptNewWorkspaceDrop: function(source, dropActor, x, y, time) {
+        this._addNewWorkspace();
+        return this._workspaces.acceptNewWorkspaceDrop(source, dropActor, x, y, time);
     }
 };
 Signals.addSignalMethods(Overview.prototype);
+
+function AddWorkspaceButton(buttonSize, buttonX, buttonY, acceptDropCallback) {
+    this._init(buttonSize, buttonX, buttonY, acceptDropCallback);
+}
+
+AddWorkspaceButton.prototype = {
+    _init: function(buttonSize, buttonX, buttonY, acceptDropCallback) {
+        this.actor = new Clutter.Group({ x: buttonX,
+                                         y: buttonY,
+                                         width: global.screen_width - buttonX,
+                                         height: global.screen_height - buttonY,
+                                         reactive: true });
+        this.actor._delegate = this;
+        this._acceptDropCallback = acceptDropCallback;
+
+        let plus = new Clutter.Texture({ x: 0,
+                                         y: 0,
+                                         width: buttonSize,
+                                         height: buttonSize });
+        plus.set_from_file(global.imagedir + 'add-workspace.svg');
+        this.actor.add_actor(plus);
+    },
+
+    // Draggable target interface
+    acceptDrop: function(source, actor, x, y, time) {
+        return this._acceptDropCallback(source, actor, x, y, time);
+    }
+};
