@@ -25,6 +25,12 @@ const WINDOWCLONE_TITLE_COLOR = new Clutter.Color();
 WINDOWCLONE_TITLE_COLOR.from_pixel(0xffffffff);
 const FRAME_COLOR = new Clutter.Color();
 FRAME_COLOR.from_pixel(0xffffffff);
+const LIGHTBOX_COLOR = new Clutter.Color();
+LIGHTBOX_COLOR.from_pixel(0x00000044);
+
+const SCROLL_SCALE_AMOUNT = 100 / 5;
+
+const ZOOM_OVERLAY_FADE_TIME = 0.15;
 
 // Define a layout scheme for small window counts. For larger
 // counts we fall back to an algorithm. We need more schemes here
@@ -40,13 +46,53 @@ const POSITIONS = {
         5: [[0.165, 0.25, 0.32], [0.495, 0.25, 0.32], [0.825, 0.25, 0.32], [0.25, 0.75, 0.32], [0.75, 0.75, 0.32]]
 };
 
+
+function _interpolate(start, end, step) {
+    return start + (end - start) * step;
+}
+
+function _clamp(value, min, max) {
+    return Math.max(min, Math.min(max, value));
+}
+
 // Spacing between workspaces. At the moment, the same spacing is used
 // in both zoomed-in and zoomed-out views; this is slightly
 // metaphor-breaking, but the alternatives are also weird.
 const GRID_SPACING = 15;
 const FRAME_SIZE = GRID_SPACING / 3;
 
-let buttonSize = false;
+function ScaledPoint(x, y, scaleX, scaleY) {
+    [this.x, this.y, this.scaleX, this.scaleY] = arguments;
+}
+
+ScaledPoint.prototype = {
+    getPosition : function() {
+        return [this.x, this.y];
+    },
+
+    getScale : function() {
+        return [this.scaleX, this.scaleY];
+    },
+
+    setPosition : function(x, y) {
+        [this.x, this.y] = arguments;
+    },
+
+    setScale : function(scaleX, scaleY) {
+        [this.scaleX, this.scaleY] = arguments;
+    },
+
+    interpPosition : function(other, step) {
+        return [_interpolate(this.x, other.x, step),
+                _interpolate(this.y, other.y, step)];
+    },
+
+    interpScale : function(other, step) {
+        return [_interpolate(this.scaleX, other.scaleX, step),
+                _interpolate(this.scaleY, other.scaleY, step)];
+    }
+};
+
 
 function WindowClone(realWindow) {
     this._init(realWindow);
@@ -61,12 +107,18 @@ WindowClone.prototype = {
         this.actor._delegate = this;
         this.realWindow = realWindow;
         this.metaWindow = realWindow.meta_window;
+        this.metaWindow._delegate = this;
         this.origX = realWindow.x;
         this.origY = realWindow.y;
 
+        this._title = null;
+
         this.actor.connect('button-release-event',
                            Lang.bind(this, this._onButtonRelease));
 
+        this.actor.connect('scroll-event',
+                           Lang.bind(this, this._onScroll));
+
         this.actor.connect('enter-event',
                            Lang.bind(this, this._onEnter));
         this.actor.connect('leave-event',
@@ -79,6 +131,18 @@ WindowClone.prototype = {
         this._inDrag = false;
     },
 
+    setVisibleWithChrome: function(visible) {
+        if (visible) {
+            this.actor.show();
+            if (this._title)
+                this._title.show();
+        } else {
+            this.actor.hide();
+            if (this._title)
+                this._title.hide();
+        }
+    },
+
     destroy: function () {
         this.actor.destroy();
         if (this._title)
@@ -104,12 +168,103 @@ WindowClone.prototype = {
             return;
 
         this._havePointer = false;
-
-        if (Tweener.isTweening(this.actor))
-            return;
-
-    	actor.raise(this.stackAbove);
         this._updateTitle();
+
+        if (this._zoomStep)
+            this._zoomEnd();
+    },
+
+    _onScroll : function (actor, event) {
+        let direction = event.get_scroll_direction();
+        if (direction == Clutter.ScrollDirection.UP) {
+            if (this._zoomStep == undefined)
+                this._zoomStart();
+            if (this._zoomStep < 100) {
+                this._zoomStep += SCROLL_SCALE_AMOUNT;
+                this._zoomUpdate();
+            }
+        } else if (direction == Clutter.ScrollDirection.DOWN) {
+            if (this._zoomStep > 0) {
+                this._zoomStep -= SCROLL_SCALE_AMOUNT;
+                this._zoomStep = Math.max(0, this._zoomStep);
+                this._zoomUpdate();
+            }
+            if (this._zoomStep <= 0.0)
+                this._zoomEnd();
+        }
+
+    },
+
+    _zoomUpdate : function () {
+        [this.actor.x, this.actor.y] = this._zoomGlobalOrig.interpPosition(this._zoomTarget, this._zoomStep / 100);
+        [this.actor.scale_x, this.actor.scale_y] = this._zoomGlobalOrig.interpScale(this._zoomTarget, this._zoomStep / 100);
+
+        let [width, height] = this.actor.get_transformed_size();
+
+        this.actor.x = _clamp(this.actor.x, 0, global.screen_width  - width);
+        this.actor.y = _clamp(this.actor.y, Panel.PANEL_HEIGHT, global.screen_height - height);
+    },
+
+    _zoomStart : function () {
+        this._zoomOverlay = new Clutter.Rectangle({ reactive: true,
+                                                    color: LIGHTBOX_COLOR,
+                                                    border_width: 0,
+                                                    x: 0,
+                                                    y: 0,
+                                                    width: global.screen_width,
+                                                    height: global.screen_height,
+                                                    opacity: 0 });
+        this._zoomOverlay.show();
+        global.stage.add_actor(this._zoomOverlay);
+        Tweener.addTween(this._zoomOverlay,
+                         { opacity: 255,
+                           time: ZOOM_OVERLAY_FADE_TIME,
+                           transition: "easeOutQuad"
+                         });
+
+        this._zoomLocalOrig  = new ScaledPoint(this.actor.x, this.actor.y, this.actor.scale_x, this.actor.scale_y);
+        this._zoomGlobalOrig = new ScaledPoint();
+        let parent = this._origParent = this.actor.get_parent();
+        [width, height] = this.actor.get_transformed_size();
+        this._zoomGlobalOrig.setPosition.apply(this._zoomGlobalOrig, this.actor.get_transformed_position());
+        this._zoomGlobalOrig.setScale(width / this.actor.width, height / this.actor.height);
+
+        this._zoomOverlay.raise_top();
+        this._zoomOverlay.show();
+
+        this.actor.reparent(global.stage);
+
+        [this.actor.x, this.actor.y]             = this._zoomGlobalOrig.getPosition();
+        [this.actor.scale_x, this.actor.scale_y] = this._zoomGlobalOrig.getScale();
+
+        this.actor.raise_top();
+
+        this._zoomTarget = new ScaledPoint(0, 0, 1.0, 1.0);
+        this._zoomTarget.setPosition(this.actor.x - (this.actor.width - width) / 2, this.actor.y - (this.actor.height - height) / 2);
+        this._zoomStep = 0;
+
+        this._hideEventId = Main.overview.connect('hiding', Lang.bind(this, function () { this._zoomEnd(); }));
+        this._zoomUpdate();
+    },
+
+    _zoomEnd : function () {
+        this.actor.reparent(this._origParent);
+
+        [this.actor.x, this.actor.y]             = this._zoomLocalOrig.getPosition();
+        [this.actor.scale_x, this.actor.scale_y] = this._zoomLocalOrig.getScale();
+
+        this._adjustTitle();
+
+        this._zoomOverlay.destroy();
+        Main.overview.disconnect(this._hideEventId);
+
+        this._zoomLocalPosition  = undefined;
+        this._zoomLocalScale     = undefined;
+        this._zoomGlobalPosition = undefined;
+        this._zoomGlobalScale    = undefined;
+        this._zoomTargetPosition = undefined;
+        this._zoomStep       = undefined;
+        this._zoomOverlay    = undefined;
     },
 
     _onButtonRelease : function (actor, event) {
@@ -231,7 +386,6 @@ DesktopClone.prototype = {
             this.actor = new Clutter.Clone({ source: window.get_texture(),
                                              reactive: true });
         } else {
-            let global = Shell.Global.get();
             this.actor = new Clutter.Rectangle({ color: global.stage.color,
                                                  reactive: true,
                                                  width: global.screen_width,
@@ -262,9 +416,6 @@ function Workspace(workspaceNum, parentActor) {
 
 Workspace.prototype = {
     _init : function(workspaceNum, parentActor) {
-        let me = this;
-        let global = Shell.Global.get();
-
         this.workspaceNum = workspaceNum;
         this._metaWorkspace = global.screen.get_workspace_by_index(workspaceNum);
 
@@ -280,6 +431,14 @@ Workspace.prototype = {
         this.actor.height = global.screen_height;
         this.scale = 1.0;
 
+        this._lightbox = new Clutter.Rectangle({ color: LIGHTBOX_COLOR });
+        this.actor.connect('notify::allocation', Lang.bind(this, function () {
+            let [width, height] = this.actor.get_size();
+            this._lightbox.set_size(width, height);
+        }));
+        this.actor.add_actor(this._lightbox);
+        this._lightbox.hide();
+
         let windows = global.get_windows().filter(this._isMyWindow, this);
 
         // Find the desktop window
@@ -311,6 +470,9 @@ Workspace.prototype = {
             }
         }
 
+        // A filter for what windows we display
+        this._showOnlyWindows = null;
+
         // Track window changes
         this._windowAddedId = this._metaWorkspace.connect('window-added',
                                                           Lang.bind(this, this._windowAdded));
@@ -326,7 +488,6 @@ Workspace.prototype = {
     },
 
     updateRemovable : function() {
-        let global = Shell.Global.get();
         let removable = (this._windows.length == 1 /* just desktop */ &&
                          this.workspaceNum != 0 &&
                          this.workspaceNum == global.screen.n_workspaces - 1);
@@ -335,8 +496,8 @@ Workspace.prototype = {
             if (this._removeButton)
                 return;
 
-            this._removeButton = new Clutter.Texture({ width: buttonSize,
-                                                       height: buttonSize,
+            this._removeButton = new Clutter.Texture({ width: Overview.addRemoveButtonSize,
+                                                       height: Overview.addRemoveButtonSize,
                                                        reactive: true
                                                      });
             this._removeButton.set_from_file(global.imagedir + "remove-workspace.svg");
@@ -381,11 +542,60 @@ Workspace.prototype = {
         return -1;
     },
 
+    /**
+     * lookupCloneForMetaWindow:
+     * @metaWindow: A #MetaWindow
+     *
+     * Given a #MetaWindow instance, find the WindowClone object
+     * which represents it in the workspaces display.
+     */
     lookupCloneForMetaWindow: function (metaWindow) {
         let index = this._lookupIndex (metaWindow);
         return index < 0 ? null : this._windows[index];
     },
 
+    containsMetaWindow: function (metaWindow) {
+        return this._lookupIndex(metaWindow) >= 0;
+    },
+
+    setShowOnlyWindows: function(showOnlyWindows, reposition) {
+        this._showOnlyWindows = showOnlyWindows;
+        this._resetCloneVisibility();
+        if (reposition)
+            this.positionWindows(false);
+    },
+
+    /**
+     * setLightboxMode:
+     * @showLightbox: If true, dim background and allow highlighting a specific window
+     *
+     * This function also resets the highlighted window state.
+     */
+    setLightboxMode: function (showLightbox) {
+        if (showLightbox) {
+            this.setHighlightWindow(null);
+            this._lightbox.show();
+        } else {
+            this._lightbox.hide();
+        }
+    },
+
+    /**
+     * setHighlightWindow:
+     * @metaWindow: A #MetaWindow
+     *
+     * Draw the user's attention to the given window @metaWindow.
+     */
+    setHighlightWindow: function (metaWindow) {
+        for (let i = 0; i < this._windows.length; i++) {
+            this._windows[i].actor.lower(this._lightbox);
+        }
+        if (metaWindow != null) {
+            let clone = this.lookupCloneForMetaWindow(metaWindow);
+            clone.actor.raise(this._lightbox);
+        }
+    },
+
     _adjustRemoveButton : function() {
         this._removeButton.set_scale(1.0 / this.actor.scale_x,
                                      1.0 / this.actor.scale_y);
@@ -402,7 +612,6 @@ Workspace.prototype = {
 
     // Mark the workspace selected/not-selected
     setSelected : function(selected) {
-        let global = Shell.Global.get();
         // Don't draw a frame if we only have one workspace
         if (selected && global.screen.n_workspaces > 1) {
             if (this._frame)
@@ -434,18 +643,51 @@ Workspace.prototype = {
                              this._desktop.actor.height + 2 * FRAME_SIZE / this.actor.scale_y);
     },
 
-    // Reposition all windows in their zoomed-to-Overview position. if workspaceZooming
-    // is true, then the workspace is moving at the same time and we need to take
-    // that into account
-    positionWindows : function(workspaceZooming) {
-        let global = Shell.Global.get();
-
+    _resetCloneVisibility: function () {
         for (let i = 1; i < this._windows.length; i++) {
             let clone = this._windows[i];
             let icon = this._windowIcons[i];
-            clone.stackAbove = this._windows[i - 1].actor;
 
-            let [xCenter, yCenter, fraction] = this._computeWindowPosition(i);
+            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows)) {
+                clone.setVisibleWithChrome(false);
+                icon.hide();
+            } else {
+                clone.setVisibleWithChrome(true);
+            }
+        }
+    },
+
+    /**
+     * positionWindows:
+     * @workspaceZooming: If true, then the workspace is moving at the same time and we need to take that into account.
+     */
+    positionWindows : function(workspaceZooming) {
+        let totalVisible = 0;
+
+        for (let i = 1; i < this._windows.length; i++) {
+            let clone = this._windows[i];
+
+            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows))
+                continue;
+
+            totalVisible += 1;
+        }
+
+        let previousWindow = this._windows[0];
+        let visibleIndex = 0;
+        for (let i = 1; i < this._windows.length; i++) {
+            let clone = this._windows[i];
+            let icon = this._windowIcons[i];
+
+            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows))
+                continue;
+
+            clone.stackAbove = previousWindow.actor;
+            previousWindow = clone;
+
+            visibleIndex += 1;
+
+            let [xCenter, yCenter, fraction] = this._computeWindowPosition(visibleIndex, totalVisible);
             xCenter = xCenter * global.screen_width;
             yCenter = yCenter * global.screen_height;
 
@@ -508,6 +750,8 @@ Workspace.prototype = {
         for (let i = 1; i < this._windows.length; i++) {
             let clone = this._windows[i];
             let icon = this._windowIcons[i];
+            if (this._showOnlyWindows != null && !(clone.metaWindow in this._showOnlyWindows))
+                continue;
             this._fadeInWindowIcon(clone, icon);
         }
     },
@@ -520,7 +764,6 @@ Workspace.prototype = {
     },
 
     _windowRemoved : function(metaWorkspace, metaWin) {
-        let global = Shell.Global.get();
         let win = metaWin.get_compositor_private();
 
         // find the position of the window in our list
@@ -672,8 +915,6 @@ Workspace.prototype = {
 
     // Animates the addition of a new (empty) workspace
     slideIn : function(oldScale) {
-        let global = Shell.Global.get();
-
         if (this.gridCol > this.gridRow) {
             this.actor.set_position(global.screen_width, this.gridY);
             this.actor.set_scale(oldScale, oldScale);
@@ -695,7 +936,6 @@ Workspace.prototype = {
     
     // Animates the removal of a workspace
     slideOut : function(onComplete) {
-        let global = Shell.Global.get();
         let destX = this.actor.x, destY = this.actor.y;
 
         this._hideAllIcons();
@@ -722,8 +962,6 @@ Workspace.prototype = {
     },
     
     destroy : function() {
-        let global = Shell.Global.get();
-
         Tweener.removeTweens(this.actor);
         this.actor.destroy();
         this.actor = null;
@@ -794,10 +1032,10 @@ Workspace.prototype = {
         return clone;
     },
 
-    _computeWindowPosition : function(index) {
+    _computeWindowPosition : function(index, totalWindows) {
         // ignore this._windows[0], which is the desktop
         let windowIndex = index - 1;
-        let numberOfWindows = this._windows.length - 1;
+        let numberOfWindows = totalWindows;
 
         if (numberOfWindows in POSITIONS)
             return POSITIONS[numberOfWindows][windowIndex];
@@ -820,7 +1058,6 @@ Workspace.prototype = {
     },
 
     _removeSelf : function(actor, event) {
-        let global = Shell.Global.get();
         let screen = global.screen;
         let workspace = screen.get_workspace_by_index(this.workspaceNum);
 
@@ -830,8 +1067,6 @@ Workspace.prototype = {
 
     // Draggable target interface
     acceptDrop : function(source, actor, x, y, time) {
-        let global = Shell.Global.get();
-
         if (source instanceof WindowClone) {
             let win = source.realWindow;
             if (this._isMyWindow(win))
@@ -862,25 +1097,25 @@ Workspace.prototype = {
 
 Signals.addSignalMethods(Workspace.prototype);
 
-function Workspaces(width, height, x, y, addButtonSize, addButtonX, addButtonY) {
-    this._init(width, height, x, y, addButtonSize, addButtonX, addButtonY);
+function Workspaces(width, height, x, y) {
+    this._init(width, height, x, y);
 }
 
 Workspaces.prototype = {
-    _init : function(width, height, x, y, addButtonSize, addButtonX, addButtonY) {
-        let global = Shell.Global.get();
-
+    _init : function(width, height, x, y) {
         this.actor = new Clutter.Group();
 
-        let screenHeight = global.screen_height;
-          
         this._width = width;
         this._height = height;
         this._x = x;
         this._y = y;
 
+        this._windowSelectionAppId = null;
+
         this._workspaces = [];
-        
+
+        this._highlightWindow = null;
+
         let activeWorkspaceIndex = global.screen.get_active_workspace_index();
         let activeWorkspace;
 
@@ -893,17 +1128,7 @@ Workspaces.prototype = {
             }
         }
         activeWorkspace.actor.raise_top();
-        this._positionWorkspaces(global);
-
-        // Save the button size as a global variable so we can us it to create 
-        // matching button sizes for workspace remove buttons.
-        buttonSize = addButtonSize;
-
-        // Create (+) button
-        this.addButton = new AddWorkspaceButton(addButtonSize, addButtonX, addButtonY, Lang.bind(this, this._addWorkspaceAcceptDrop));
-        this.addButton.actor.connect('button-release-event', Lang.bind(this, this._appendNewWorkspace));
-        this.actor.add_actor(this.addButton.actor);
-        this.addButton.actor.lower_bottom();
+        this._positionWorkspaces();
 
         let lastWorkspace = this._workspaces[this._workspaces.length - 1];
         lastWorkspace.updateRemovable(true);
@@ -927,6 +1152,14 @@ Workspaces.prototype = {
                                           Lang.bind(this, this._activeWorkspaceChanged));
     },
 
+    _lookupWorkspaceForMetaWindow: function (metaWindow) {
+        for (let i = 0; i < this._workspaces.length; i++) {
+            if (this._workspaces[i].containsMetaWindow(metaWindow))
+                return this._workspaces[i];
+        }
+        return null;
+    },
+
     _lookupCloneForMetaWindow: function (metaWindow) {
         for (let i = 0; i < this._workspaces.length; i++) {
             let clone = this._workspaces[i].lookupCloneForMetaWindow(metaWindow);
@@ -936,9 +1169,64 @@ Workspaces.prototype = {
         return null;
     },
 
-    // Should only be called from active Overview context
-    activateWindowFromOverview: function (metaWindow, time) {
-        let global = Shell.Global.get();
+    setHighlightWindow: function (metaWindow) {
+        // Looping over all workspaces is easier than keeping track of the last
+        // highlighted window while trying to handle the window or workspace possibly
+        // going away.
+        for (let i = 0; i < this._workspaces.length; i++) {
+            this._workspaces[i].setHighlightWindow(null);
+        }
+        if (metaWindow != null) {
+            let workspace = this._lookupWorkspaceForMetaWindow(metaWindow);
+            workspace.setHighlightWindow(metaWindow);
+        }
+    },
+
+    _clearApplicationWindowSelection: function(reposition) {
+        if (this._windowSelectionAppId == null)
+            return;
+        this._windowSelectionAppId = null;
+
+        for (let i = 0; i < this._workspaces.length; i++) {
+            this._workspaces[i].setLightboxMode(false);
+            this._workspaces[i].setShowOnlyWindows(null, reposition);
+        }
+    },
+
+    /**
+     * setApplicationWindowSelection:
+     * @appid: Application identifier string
+     *
+     * Enter a mode which shows only the windows owned by the
+     * given application, and allow highlighting of a specific
+     * window with setHighlightWindow().
+     */
+    setApplicationWindowSelection: function (appId) {
+        if (appId == null) {
+            this._clearApplicationWindowSelection(true);
+            return;
+        }
+
+        if (appId == this._windowSelectionAppId)
+            return;
+
+        this._windowSelectionAppId = appId;
+
+        let appSys = Shell.AppMonitor.get_default();
+
+        let showOnlyWindows = {};
+        let windows = appSys.get_windows_for_app(appId);
+        for (let i = 0; i < windows.length; i++) {
+            showOnlyWindows[windows[i]] = 1;
+        }
+
+        for (let i = 0; i < this._workspaces.length; i++) {
+            this._workspaces[i].setLightboxMode(true);
+            this._workspaces[i].setShowOnlyWindows(showOnlyWindows, true);
+        }
+    },
+
+    _activateWindowInternal: function (metaWindow, time) {
         let activeWorkspaceNum = global.screen.get_active_workspace_index();
         let windowWorkspaceNum = metaWindow.get_workspace().index();
 
@@ -953,12 +1241,27 @@ Workspaces.prototype = {
         }
     },
 
+    /**
+     * activateWindowFromOverview:
+     * @metaWindow: A #MetaWindow
+     * @time: Integer even timestamp
+     *
+     * This function exits the overview, switching to the given @metaWindow.
+     * If an application filter is in effect, it will be cleared.
+     */
+    activateWindowFromOverview: function (metaWindow, time) {
+        if (this._windowSelectionAppId != null) {
+            this._clearApplicationWindowSelection(false);
+        }
+        this._activateWindowInternal(metaWindow, time);
+        Main.overview.hide();
+    },
+
     hide : function() {
-        let global = Shell.Global.get();
         let activeWorkspaceIndex = global.screen.get_active_workspace_index();
         let activeWorkspace = this._workspaces[activeWorkspaceIndex];
 
-        this._positionWorkspaces(global);
+        this._positionWorkspaces();
         activeWorkspace.actor.raise_top();
 
         for (let w = 0; w < this._workspaces.length; w++)
@@ -966,8 +1269,6 @@ Workspaces.prototype = {
     },
 
     destroy : function() {
-        let global = Shell.Global.get();
-
         for (let w = 0; w < this._workspaces.length; w++)
             this._workspaces[w].destroy();
         this._workspaces = [];
@@ -985,7 +1286,6 @@ Workspaces.prototype = {
 
     // Get the grid position of the active workspace.
     getActiveWorkspacePosition : function() {
-        let global = Shell.Global.get();
         let activeWorkspaceIndex = global.screen.get_active_workspace_index();
         let activeWorkspace = this._workspaces[activeWorkspaceIndex];
 
@@ -1002,7 +1302,7 @@ Workspaces.prototype = {
     // first row.)
     //
     // FIXME: need to make the metacity internal layout agree with this!
-    _positionWorkspaces : function(global) {
+    _positionWorkspaces : function() {
         let gridWidth = Math.ceil(Math.sqrt(this._workspaces.length));
         let gridHeight = Math.ceil(this._workspaces.length / gridWidth);
 
@@ -1040,8 +1340,6 @@ Workspaces.prototype = {
     },
 
     _workspacesChanged : function() {
-        let global = Shell.Global.get();
-
         let oldNumWorkspaces = this._workspaces.length;
         let newNumWorkspaces = global.screen.n_workspaces;
 
@@ -1074,7 +1372,7 @@ Workspaces.prototype = {
         newLastWorkspace.updateRemovable();
 
         // Figure out the new layout
-        this._positionWorkspaces(global);
+        this._positionWorkspaces();
         let newScale = this._workspaces[0].scale;
         let newGridWidth = Math.ceil(Math.sqrt(newNumWorkspaces));
         let newGridHeight = Math.ceil(newNumWorkspaces / newGridWidth);
@@ -1125,42 +1423,13 @@ Workspaces.prototype = {
         this.actor.add_actor(workspace.actor);
     },
 
-    _appendNewWorkspace : function() {
-        let global = Shell.Global.get();
-
-        global.screen.append_new_workspace(false, global.screen.get_display().get_current_time());
-    },
-
-    // Creates a new workspace and drops the dropActor there
-    _addWorkspaceAcceptDrop : function(source, dropActor, x, y, time) {
-        this._appendNewWorkspace();
+    // Handles a drop onto the (+) button; assumes the new workspace
+    // has already been added
+    acceptNewWorkspaceDrop : function(source, dropActor, x, y, time) {
         return this._workspaces[this._workspaces.length - 1].acceptDrop(source, dropActor, x, y, time);
     }
 };
 
-function AddWorkspaceButton(buttonSize, buttonX, buttonY, acceptDropCallback) {
-    this._init(buttonSize, buttonX, buttonY, acceptDropCallback);
-}
-
-AddWorkspaceButton.prototype = {
-    _init : function(buttonSize, buttonX, buttonY, acceptDropCallback) {
-        this.actor = new Clutter.Texture({ x: buttonX,
-                                           y: buttonY,
-                                           width: buttonSize,
-                                           height: buttonSize,
-                                           reactive: true });
-        this._acceptDropCallback = acceptDropCallback;
-        let global = Shell.Global.get();
-        this.actor._delegate = this;
-        this.actor.set_from_file(global.imagedir + 'add-workspace.svg');
-    },
-
-    // Draggable target interface
-    acceptDrop : function(source, actor, x, y, time) {
-        return this._acceptDropCallback(source, actor, x, y, time);
-    }
-};
-
 // Create a SpecialPropertyModifier to let us move windows in a
 // straight line on the screen even though their containing workspace
 // is also moving.
