@@ -36,10 +36,6 @@ GType            shell_global_get_type            (void) G_GNUC_CONST;
 gboolean shell_clutter_texture_set_from_pixbuf (ClutterTexture *texture,
                                                 GdkPixbuf      *pixbuf);
 
-guint16 shell_get_event_key_symbol(ClutterEvent *event);
-
-guint16 shell_get_button_event_click_count(ClutterEvent *event);
-
 ClutterActor *shell_get_event_related(ClutterEvent *event);
 
 ShellGlobal *shell_global_get (void);
